package packages.ui;

import java.util.ArrayList;
import java.util.List;

import greenfoot.*;
import packages.event.Action;
/**
 * This is a Button class. Use it for interfaces. The Button works completly
 * automaticly, you do not have to feed it any information after the
 * construction. It offers selveral methods to get information about
 * the users interactions with it. It may also go to a certain world when clicked
 * to move through menus.
 * 
 * @author RcCookie
 * @version 1.2
 */
public class Button extends Actor {
    /**
     * Time between repaints of the button, in fraction of a second. Used to increase performance.
     * <p><b>Important:</b> This does not include the painting of hover and click
     * animations, it only effects the repaint rate of changes in images and texts.
     * <p>Default is 1/60: Enough to paint every frame on a 60Hz screen.
     */
    public static double TIME_BETWEEN_REPAINTS = 1 / 60d;

    protected static final Color DEF_BACKGROUND_COLOR = Color.GRAY;
    protected static final Color OUTLINE_COLOR = new Color(35, 35, 35);
    protected static final Color CLICK_COLOR = new Color(0, 0, 0, 80);
    protected static final int MIN_BORDER = 3;
    protected static final int MIN_X = 10, MIN_Y = 10;
    protected static final double HOVER_SCALE = 1.05;

    // ------------------------------------------
    // Button properties
    // ------------------------------------------

    /**
     * The image given as input, or null if a color was chosen.
     */
    protected GreenfootImage background;

    /**
     * The text object representing all text properties.
     */
    protected Text text;

    /**
     * The minimum x and y size of the button.
     */
    protected int minX, minY;

    /**
     * Weather an outline should be drawn.
     */
    protected boolean outline;

    /**
     * Sets weather the button is enabled.
     */
    public boolean enabled = true;


    /**
     * The time stamp of the last executed repaint.
     */
    private long lastRepaint = System.nanoTime();

    /**
     * Weather this button should always be repainted. This will have an 
     * impact on the performance.
     */
    public boolean alwaysRepaint = false;



    // ------------------------------------------
    // Button states
    // ------------------------------------------

    /**
     * The number of times this button has been clicked.
     */
    protected int clickCount;

    /**
     * The number of frames this button was held down.
     */
    protected int pressTime;

    /**
     * Weather the button is touched by the mouse right now.
     */
    protected boolean touched;

    /**
     * Weather the button is pressed down by the mouse right now.
     */
    protected boolean pressed;

    /**
     * The actions to execute when the button was clicked.
     */
    protected final List<Action> clickActions = new ArrayList<>();
    
    /**
     * The actions to execute when the button was pressed.
     */
    protected final List<Action> pressActions = new ArrayList<>();
    
    /**
     * The actions to execute when the button was released.
     */
    protected final List<Action> releaseActions = new ArrayList<>();


    private final Action textUpdateAction;



    // ------------------------------------------
    // Cashing
    // ------------------------------------------

    /**
     * The standart image of the button.
     */
    protected GreenfootImage image;

    /**
     * The buttons image when hovered over it with the mouse.
     */
    protected GreenfootImage hoveredImage;

    /**
     * The buttons image when it is being clicked.
     */
    protected GreenfootImage clickedImage;










    //--------------------------------------------------
    // String constructors
    //--------------------------------------------------


    

    /**
     * Constructs a new grey Button with the given title.
     * 
     * @param title The text printed onto the button
     */
    public Button(String title){
        this(new Text(title));
    }
    
    /**
     * Constructs a new grey Button with the given title and a the given width
     * and height.
     * 
     * @param title The text printed onto the button
     * @param minX The width of the button
     * @param minY The height of the button
     */
    public Button(String title, int minX, int minY){
        this(new Text(title), minX, minY);
    }


    //--------------------------------------------------
    // Plain button constructors
    //--------------------------------------------------


    /**
     * Constructs a new gray button with the given size.
     * <p>The size is labled as <i>minimum</i> because later changes to the text may change
     * the required size.
     * 
     * @param minX The width of the button
     * @param minY The height of the button
     */
    public Button(int minX, int minY) {
        this(minX, minY, DEF_BACKGROUND_COLOR);
    }

    /**
     * Constructs a new button in the given size and color.
     * <p>The size is labled as <i>minimum</i> because later changes to the text may change
     * the required size.
     * 
     * @param minX The width of the button
     * @param minY The height of the button
     * @param backgroundColor The color of the button
     */
    public Button(int minX, int minY, Color backgroundColor) {
        this(minX, minY, backgroundColor, true);
    }

    /**
     * Constructs a new button in the given size and color and if chosen with a dark outline.
     * <p>The size is labled as <i>minimum</i> because later changes to the text may change
     * the required size.
     * 
     * @param minX The width of the button
     * @param minY The height of the button
     * @param backgroundColor The color of the button
     * @param outline Weather a dark outline should be drawn
     */
    public Button(int minX, int minY, Color backgroundColor, boolean outline) {
        this(coloredText(backgroundColor), minX, minY, outline);
    }



    //--------------------------------------------------
    // Image constructors
    //--------------------------------------------------


    /**
     * Constructs a new button with the size of the given image and that image as background.
     * <p>If {@code background} is {@code null}, the background will be filled with a gray background
     * color.
     * <p>The given image will not be cloned so changes to that instance after passing it into this button
     * will have an effect. HOWEVER, the image WILL NOT be updated automaticly by default. There are multiple
     * ways to update the buttons image though:
     * <ul>
     * <li>The button image gets updates for other reasons automatically (for example due to a change in text)
     * <li>Calling {@code createAndSetImages()} manually after adjusting the image.
     * <li>Setting the image using {@code setBackground(GreenfootImage)} (even though it is the same instance)
     * <li>Setting {@code alwaysRepaint} to {@code true}. This will automatically update the image every frame
     * </ul>
     * <p>It is not reccomended to always simply set {@code alwaysRepaint} to {@code true} as repainting takes
     * a long time and has a considerable impact on performance.
     * 
     * @param background The image for the button's background
     */
    public Button(GreenfootImage background) {
        this(background, false);
    }

    /**
     * Constructs a new button with the size of the given image and that image as background.
     * If chosen a dark outline will be drawn.
     * <p>If {@code background} is {@code null}, the background will be filled with a gray background
     * color.
     * <p>The given image will not be cloned so changes to that instance after passing it into this button
     * will have an effect. HOWEVER, the image WILL NOT be updated automaticly by default. There are multiple
     * ways to update the buttons image though:
     * <ul>
     * <li>The button image gets updates for other reasons automatically (for example due to a change in text)
     * <li>Calling {@code createAndSetImages()} manually after adjusting the image.
     * <li>Setting the image using {@code setBackground(GreenfootImage)} (even though it is the same instance)
     * <li>Setting {@code alwaysRepaint} to {@code true}. This will automatically update the image every frame
     * </ul>
     * <p>It is not reccomended to always simply set {@code alwaysRepaint} to {@code true} as repainting takes
     * a long time and has a considerable impact on performance.
     * 
     * @param background The image for the button's background
     * @param Weather a dark outline should be drawn
     */
    public Button(GreenfootImage background, boolean outline) {
        this(coloredText(new Color(0, 0, 0, 0)), background, outline);
    }

    private static final Text coloredText(Color color) {
        return new Text(null, 1, Color.BLACK, color);
    }




    //--------------------------------------------------
    // Text constructors
    //--------------------------------------------------




    /**
     * Constructs a new button from the given text. Content and size
     * are updated every frame.
     * 
     * @param text The text object the button should be based on
     */
    public Button(Text text){
        this(text, true);
    }

    /**
     * Constructs a new button with the given text and when chosen a dark outline.
     * 
     * @param text The text object for the button
     * @param outline Weather a dark outline should be drawn
     */
    public Button(Text text, boolean outline) {
        this(text, MIN_X, MIN_Y, outline);
    }
    
    /**
     * Constructs a new button from the given text that is at least
     * as big as inputed. Content and size are updated every frame.
     * 
     * @param text The text object the button should be based on
     * @param minX The minimum width of the button
     * @param minY The minimum height of the button
     */
    public Button(Text text, int minX, int minY){
        this(text, minX, minY, null);
    }

    /**
     * Constructs a new button that is at least a big as the given image. If the text is bigger than
     * the image, the text's background color will be used to fill the gap.
     * <p>If {@code background} is {@code null}, the background will be filled with the text's background
     * color.
     * <p>The given image will not be cloned so changes to that instance after passing it into this button
     * will have an effect. HOWEVER, the image WILL NOT be updated automaticly by default. There are multiple
     * ways to update the buttons image though:
     * <ul>
     * <li>The button image gets updates for other reasons automatically (for example due to a change in text)
     * <li>Calling {@code createAndSetImages()} manually after adjusting the image.
     * <li>Setting the image using {@code setBackground(GreenfootImage)} (even though it is the same instance)
     * <li>Setting {@code alwaysRepaint} to {@code true}. This will automatically update the image every frame
     * </ul>
     * <p>It is not reccomended to always simply set {@code alwaysRepaint} to {@code true} as repainting takes
     * a long time and has a considerable impact on performance.
     * 
     * @param text The text object for the button
     * @param background The background image for the button
     */
    public Button(Text text, GreenfootImage background) {
        this(text, background, true);
    }

    /**
     * Constructs a new button that is at least a big as the given image. If the text is bigger than
     * the image, the text's background color will be used to fill the gap.
     * <p>If {@code background} is {@code null}, the background will be filled with the text's background
     * color.
     * <p>The given image will not be cloned so changes to that instance after passing it into this button
     * will have an effect. HOWEVER, the image WILL NOT be updated automaticly by default. There are multiple
     * ways to update the buttons image though:
     * <ul>
     * <li>The button image gets updates for other reasons automatically (for example due to a change in text)
     * <li>Calling {@code createAndSetImages()} manually after adjusting the image.
     * <li>Setting the image using {@code setBackground(GreenfootImage)} (even though it is the same instance)
     * <li>Setting {@code alwaysRepaint} to {@code true}. This will automatically update the image every frame
     * </ul>
     * <p>It is not reccomended to always simply set {@code alwaysRepaint} to {@code true} as repainting takes
     * a long time and has a considerable impact on performance.
     * 
     * @param text The text object for the button
     * @param background The background image for the button
     * @param outline Weather a dark outline should be drawn
     */
    public Button(Text text, GreenfootImage background, boolean outline) {
        this(text, background != null ? background.getWidth() : MIN_X, background != null ? background.getHeight() : MIN_Y, background, outline);
    }

    /**
     * Constructs a new button from the given text that is at least
     * as big as inputed. Content and size are updated every frame.
     * <p>The given image will not be cloned so changes to that instance after passing it into this button
     * will have an effect. HOWEVER, the image WILL NOT be updated automaticly by default. There are multiple
     * ways to update the buttons image though:
     * <ul>
     * <li>The button image gets updates for other reasons automatically (for example due to a change in text)
     * <li>Calling {@code createAndSetImages()} manually after adjusting the image.
     * <li>Setting the image using {@code setBackground(GreenfootImage)} (even though it is the same instance)
     * <li>Setting {@code alwaysRepaint} to {@code true}. This will automatically update the image every frame
     * </ul>
     * <p>It is not reccomended to always simply set {@code alwaysRepaint} to {@code true} as repainting takes
     * a long time and has a considerable impact on performance.
     * 
     * @param text The text object the button should be based on
     * @param minX The minimum width of the button
     * @param minY The minimum height of the button
     * @param background The background image for the button. If {@code null},
     *                   the background color of the text will be used
     */
    public Button(Text text, int minX, int minY, GreenfootImage background) {
        this(text, minX, minY, background, true);
    }
    
    /**
     * Constructs a new button from the given text that is at least
     * as big as inputed. It will switch to the given world unless it
     * is null. Content and size are updated every frame.
     * 
     * @param text The text object the button should be based on
     * @param minX The minimum width of the button
     * @param minY The minimum height of the button
     * @param outline If a dark outline should be drawed around the button
     * @param onClick The world to switch to when clicked
     */
    public Button(Text text, int minX, int minY, boolean outline){
        this(text, minX, minY, null, outline);
    }

    /**
     * Constructs a new Button with the specified properties.
     * <p>The given image will not be cloned so changes to that instance after passing it into this button
     * will have an effect. HOWEVER, the image WILL NOT be updated automaticly by default. There are multiple
     * ways to update the buttons image though:
     * <ul>
     * <li>The button image gets updates for other reasons automatically (for example due to a change in text)
     * <li>Calling {@code createAndSetImages()} manually after adjusting the image.
     * <li>Setting the image using {@code setBackground(GreenfootImage)} (even though it is the same instance)
     * <li>Setting {@code alwaysRepaint} to {@code true}. This will automatically update the image every frame
     * </ul>
     * <p>It is not reccomended to always simply set {@code alwaysRepaint} to {@code true} as repainting takes
     * a long time and has a considerable impact on performance.
     * 
     * @param text The text object for the button
     * @param minX The minimum width of the button
     * @param minY The minimum height of the button
     * @param background The background image for the button. If {@code null},
     *                   the background color of the text will be used
     * @param outline Weather a dark outline should be drawn
     */
    public Button(Text text, int minX, int minY, GreenfootImage background, boolean outline) {
        this.text = text;
        this.minX = minX;
        this.minY = minY;
        this.background = background;
        this.outline = outline;

        textUpdateAction = info -> createAndSetImages();
        text.addUpdateAction(textUpdateAction);

        touched = false;
        pressed = false;
        clickCount = 0;
        pressTime = 0;

        createAndSetImages();
        updateAnimations();
        setImage(image);
    }

















    /**
     * Creates the image, hoveredImage and clickedImage for the button using
     * the latest settings and aplies the matching image to the button.
     */
    protected void createAndSetImages(){

        // Create correct sized image
        int x = this.minX, y = this.minY;
        if(x < text.getImage().getWidth() + 2 * MIN_BORDER) x = text.getImage().getWidth() + 2 * MIN_BORDER;
        if(y < text.getImage().getHeight()) y = text.getImage().getHeight();
        image = new GreenfootImage(x, y);

        // Fill background around text with text background color or default color, if null
        if(background == null || background.getWidth() < x || background.getHeight() < y) {
            if(text.getBackgroundColor() == null) {
                image.setColor(DEF_BACKGROUND_COLOR);
                image.fill();
            }
            else {
                image.setColor(text.getBackgroundColor());
                int delta = x - text.getImage().getWidth();
                image.fillRect(0, 0, delta / 2, image.getHeight());
                image.fillRect(image.getWidth() - (delta - delta / 2), 0, delta / 2, image.getHeight());
                if(y > text.getImage().getHeight()){
                    int deltaX = x - text.getImage().getWidth();
                    int deltaY = y - text.getImage().getHeight();
                    image.fillRect(deltaX / 2, 0, text.getImage().getWidth(), deltaY / 2);
                    image.fillRect(deltaX / 2, image.getHeight() - (deltaY - deltaY / 2), text.getImage().getWidth(), deltaY / 2);
                }
            }
        }

        // Draw image onto background if not null
        if(background != null) {
            image.drawImage(background, x / 2 - background.getWidth() / 2, y / 2 - background.getHeight() / 2);
        }

        // Draw text onto background
        image.drawImage(text.getImage(), (x - text.getImage().getWidth()) / 2, (y - text.getImage().getHeight()) / 2);

        // Draw outline if needed
        if(outline) {
            image.setColor(OUTLINE_COLOR);
            image.drawRect(0, 0, x - 1, y - 1);
        }


        // Modified images


        // Create hover image
        hoveredImage = new GreenfootImage(image);
        hoveredImage.scale((int)(image.getWidth() * HOVER_SCALE), (int)(image.getHeight() * HOVER_SCALE));

        // Create clicked image
        clickedImage = new GreenfootImage(image);
        clickedImage.setColor(CLICK_COLOR);
        //for(int i=0; i<clickedImage.getWidth(); i++) for(int j=0; j<clickedImage.getHeight(); j++) {
            //if(clickedImage.getColorAt(i, j).getAlpha() != 0) clickedImage.fillRect(i, j, 1, 1);
        //}
        clickedImage.fill();

        updateAnimations();
    }




    /**
     * Sets the image of the button to the currently matching animation state.
     */
    protected void updateAnimations() {
        if(touched){
            if(pressed) setImage(clickedImage);
            else setImage(hoveredImage);
        }
        else setImage(image);
    }

    /**
     * Analyses the mouse interactions each frame. Do not remove
     * any code here! Use the run method instead!
     * 
     * @deprecated If this method is not executed, the button will not work!
     */
    public void act()
    {
        touched = touching();
        if(touched&&Greenfoot.mousePressed(null)){
            pressed=true;
            if(enabled) onPress();
        }
        else if(Greenfoot.mouseClicked(null)){
            if(pressed&&touched){
                clickCount++;
                if(enabled) onClick();
                if(enabled) onRelease();
                pressed = false;
            }
            else if(pressed){
                pressed = false;
                if(enabled) onRelease();

            }
        }

        
        // Regenerating images when neccecary
        long current = System.nanoTime();
        if(alwaysRepaint || (current - lastRepaint) / 1000000000d > TIME_BETWEEN_REPAINTS) {
            //createAndSetImages();
            lastRepaint = current;
        }

        // Repainting animations
        updateAnimations();
        
        if(pressed)pressTime++;
        run();
    }


    /**
     * Simulates clicking the button for the given time. The button will not
     * actually animate the click, however, click count and press time will
     * be updated and {@code onPress()}, {@code onClick()} and {@code onRelease()}
     * will be executed.
     * 
     * @param time The number of frames the button should (virtually) be pressed down.
     */
    public void click(int time){
        if(enabled) onPress();
        pressTime += time >= 0 ? time : 1;
        clickCount++;
        if(enabled) onClick();
        if(enabled) onRelease();
    }

    /**
     * Simulates clicking the button for a single frame. The button will not
     * actually animate the click, however, click count and press time will
     * be updated and {@code onPress()}, {@code onClick()} and {@code onRelease()}
     * will be executed.
     */
    public void click(){
        click(1);
    }


    /**
     * Checkes weather the mouse is touching the button right now. Is rather
     * cpu-intense so don't use it more than once per frame.
     * 
     * @return Wheather the mouse is touching the button
     */
    public boolean touching(){
        try{
            MouseInfo mouse = Greenfoot.getMouseInfo();
            return ActorVisitor.containsPoint(this, mouse.getX(), mouse.getY());
        }catch(Exception e){}
        return false;
    }


    /**
     * Executed once per frame as a replacement of the act method. Override
     * it to use.
     */
    public void run(){}

    /**
     * Executed whenever the mouse is first pressed onto the button
     */
    protected void onPress(){
        for(Action action : pressActions) action.run(null);
    }

    /**
     * Executed whenever the mouse is released after holding down the button,
     * also if is is not on it no more.
     */
    protected void onRelease(){
        for(Action action : releaseActions) action.run(null);
    }

    /**
     * Executed ehenever the mouse is released after holding down the button,
     * but only when the mouse touched the button while being released.
     */
    protected void onClick(){
        for(Action action : clickActions) action.run(null);
    }
    

    /**
     * Resets the stats of the button.
     */
    public void reset(){
        clickCount = 0;
        pressTime = 0;
    }
    
    /**
     * Sets the count of the button to the given value.
     * 
     * @param n The new value
     */
    public void setCount(int n){
        clickCount = n;
    }
    
    /**
     * Returns the number of times the button was clicked since the last
     * reset.
     * 
     * The Button is considered to be clicked when the mouse was pressed on
     * the button and then released on it again. It still counts as a click
     * if the mouse was dragged outside of the button while being pressed
     * but is released on the button.
     * 
     * @return The number of clicks
     */
    public int getClickCount(){
        return clickCount;
    }
    
    /**
     * Returns true if the button was clicked since the last reset.
     * 
     * The Button is considered to be clicked when the mouse was pressed on
     * the button and then released on it again. It still counts as a click
     * if the mouse was dragged outside of the button while being pressed
     * but is released on the button.
     * 
     * @return A boolean if the button was ever clicked at least one time
     */
    public boolean clicked(){
        return clickCount>0;
    }
    
    /**
     * Returns if the button is being pressed right now.
     * 
     * The Button is considered to be pressed when the mouse was pressed on
     * the buttonand since than holded down. It still counts as pressed
     * if the mouse was dragged outside of the button while being pressed.
     * 
     * @return If the button is pressed right now
     */
    public boolean pressed(){
        return pressed;
    }
    
    /**
     * Returns if the button is being touched by the mouse right now.
     * 
     * @return If the button is touched right now
     */
    public boolean touched(){
        return touched;
    }
    
    /**
     * Returns the number of frames the button was pressed since the last
     * reset.
     * 
     * It is considered as pressed if the boolean pressed() is true.
     * 
     * @return the time the button was pressed
     */
    public int getTimePressed(){
        return pressTime;
    }
    
    /**
     * Returns the title written onto the button.
     * 
     * @return The buttons text
     */
    public String getTitle(){
        return text.getContent();
    }

    /**
     * Returns the text object used in this button
     * 
     * @return The text object of this button
     */
    public Text getText(){
        return text;
    }


    /**
     * Override the title of the button.
     * 
     * @param title The new title
     */
    public void setTitle(String title){
        text.setContent(title);
        createAndSetImages();
    }

    /**
     * Sets the image of the button to the given one. If image is null,
     * the latest color will be used.
     * 
     * @param background The new image
     */
    public void setBackground(GreenfootImage background){
        this.background = background;
        createAndSetImages();
    }

    /**
     * Sets the text object the button is based on to the given one.
     * 
     * @param text The text object to set to
     */
    public void setText(Text text){
        if(text == null || this.text == text) return;
        this.text.removeUpdateAction(textUpdateAction);
        text.addUpdateAction(textUpdateAction);
        this.text = text;
        createAndSetImages();
    }

    /**
     * Sets weather the outline should be drawn or not
     * 
     * @param outline Weather the outline should be drawn
     */
    public void useOutline(boolean outline){
        this.outline = outline;
        createAndSetImages();
    }


    /**
     * Sets the images displayed to custom ones to allow more customisation.
     * 
     * @param image The image to be showm normally
     * @param hoveredImage The image to be showm when the mouse hovers over the button
     * @param clickedImage The image to be shown when the button is pressed down
     */
    public void setCustomImages(GreenfootImage image, GreenfootImage hoveredImage, GreenfootImage clickedImage){
        this.image = image;
        this.hoveredImage = hoveredImage;
        this.clickedImage = clickedImage;
        updateAnimations();
    }







    protected List<Action> getClickActions() {
        return clickActions;
    }

    protected List<Action> getPressActions() {
        return pressActions;
    }

    protected List<Action> getReleaseActions() {
        return releaseActions;
    }

    /**
     * Adds an action that will be executed whenever the button was clicked.
     * <p><b>Example:<b>
     * <p>{@code Button b = new Button("Hello World!");}
     * <p>{@code b.addClickAction(info -> System.out.println("Hello!"));}
     * 
     * @param action The action to add
     */
    public Button addClickAction(Action action) {
        if(action == null) return this;
        clickActions.add(action);
        return this;
    }
    
    /**
     * Adds an action that will be executed whenever the button was pressed.
     * <p><b>Example:<b>
     * <p>{@code Button b = new Button("Hello World!");}
     * <p>{@code b.addPressAction(info -> System.out.println("Hello!"));}
     * 
     * @param action The action to add
     */
    public Button addPressAction(Action action) {
        if(action == null) return this;
        pressActions.add(action);
        return this;
    }
    
    /**
     * Adds an action that will be executed whenever the button was released.
     * <p><b>Example:<b>
     * <p>{@code Button b = new Button("Hello World!");}
     * <p>{@code b.addReleaseAction(info -> System.out.println("Hello!"));}
     * 
     * @param action The action to add
     */
    public Button addReleaseAction(Action action) {
        if(action == null) return this;
        releaseActions.add(action);
        return this;
    }


    /**
     * Removes the given action from those that are ran whenever the button is clicked.
     * 
     * @param action The action to remove
     */
    public void removeClickAction(Action action) {
        clickActions.remove(action);
    }

    /**
     * Removes the given action from those that are ran whenever the button is pressed.
     * 
     * @param action The action to remove
     */
    public void removePressAction(Action action) {
        pressActions.remove(action);
    }

    /**
     * Removes the given action from those that are ran whenever the button is released.
     * 
     * @param action The action to remove
     */
    public void removereleaseAction(Action action) {
        releaseActions.remove(action);
    }
}